@@ -3,26 +3,23 @@
 #
 # Copyright © 2015-2016 Alexandre Detiste <alexandre@detiste.be>
 # Copyright © 2016 Simon McVittie <smcv@debian.org>
-#
-# This program is free software; you can redistribute it and/or
-# modify it under the terms of the GNU General Public License
-# as published by the Free Software Foundation; either version 2
-# of the License, or (at your option) any later version.
-#
-# This program is distributed in the hope that it will be useful,
-# but WITHOUT ANY WARRANTY; without even the implied warranty of
-# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
-#
-# You can find the GPL license text on a Debian system under
-# /usr/share/common-licenses/GPL-2.
+# SPDX-License-Identifier: GPL-2.0-or-later
+
+from __future__ import annotations
 
 import logging
 import os
 import subprocess
 import time
-from distutils.version import LooseVersion as Version
+from collections.abc import (Iterable)
+from typing import (TYPE_CHECKING)
 
-from . import (PackagingSystem)
+if TYPE_CHECKING:
+    from . import (PerPackageState)
+    from ..data import (Package, PackageRelation)
+    from ..game import (GameData)
+
+from . import (Compression, PackagingSystem)
 from ..util import (
         check_output,
         normalize_permissions,
@@ -31,24 +28,36 @@ from ..util import (
 
 logger = logging.getLogger(__name__)
 
+
 class RpmPackaging(PackagingSystem):
     INSTALL_CMD = ['rpm', '-U']
     CHECK_CMD = 'rpmlint'
+    BUILD_DEP = {'rpmbuild'}
     ARCH_DECODE = {
                   'all': 'noarch',
                   'i386': 'i686',
                   'amd64': 'x86_64',
                   }
+    ARCH_ENCODE = {
+        'noarch': 'all',
+        'i586': 'i386',
+        'i686': 'i386',
+        'x86_64': 'amd64',
+    }
 
-    def __init__(self, distro=None):
-        super(RpmPackaging, self).__init__()
+    def __init__(
+        self,
+        distro: str | None = None,
+        architecture: str | None = None,
+    ) -> None:
+        super(RpmPackaging, self).__init__(architecture=architecture)
         self.distro = distro
         if distro is None or distro == 'generic':
             self._contexts = ('rpm', 'generic')
         else:
             self._contexts = (distro, 'rpm', 'generic')
 
-    def is_installed(self, package):
+    def is_installed(self, package: str) -> bool:
         try:
             return 0 == subprocess.call(['rpm', '-q', package],
                                         stdout=subprocess.DEVNULL,
@@ -56,24 +65,36 @@ class RpmPackaging(PackagingSystem):
         except FileNotFoundError:
             return False
 
-    def current_version(self, package):
+    def current_version(self, package: str) -> str | None:
         try:
-            return check_output(['rpm', '-q',
-              '--qf', '%{VERSION}', package], universal_newlines=True)
+            v = check_output(
+                ['rpm', '-q', '--qf', '%{VERSION}', package],
+                text=True,
+            )
+            assert type(v) is str
+            return v
         except FileNotFoundError:
             return None
         except subprocess.CalledProcessError:
             return None
+        else:
+            return None
 
-    def is_available(self, package):
+    def is_available(self, package: str) -> bool:
         # assume no apt-like system in this base class
         return self.is_installed(package)
 
-    def available_version(self, package):
+    def available_version(self, package: str) -> str | None:
         # assume no apt-like system in this base class
         return self.current_version(package)
 
-    def install_packages(self, rpms, method=None, gain_root='su'):
+    def install_packages(
+        self,
+        rpms: Iterable[str],
+        method: str | None = None,
+        gain_root: str = 'su',
+        force: bool = False,
+    ) -> None:
         """Install one or more packages (a list of filenames)."""
 
         if not method:
@@ -91,9 +112,11 @@ class RpmPackaging(PackagingSystem):
                                 ' using rpm instead') % method)
             run_as_root(['rpm', '-U'] + list(rpms), gain_root)
 
-    def format_relation(self, pr):
+    def format_relation(self, pr: PackageRelation) -> str:
         assert not pr.contextual
         assert not pr.alternatives
+        package = pr.package
+        assert package is not None, str(pr)
 
         if pr.version is not None:
             op = pr.version_operator
@@ -102,13 +125,22 @@ class RpmPackaging(PackagingSystem):
                 op = op[0]
 
             # foo >= 1.0
-            return '%s %s %s' % (self.rename_package(pr.package), op,
-                    pr.version)
+            return '%s %s %s' % (
+                self.rename_package(package), op, pr.version,
+            )
 
-        return self.rename_package(pr.package)
+        return self.rename_package(package)
 
-    def __fill_dest_dir_rpm(self, game, package, workdir, destdir, compress,
-            architecture, release):
+    def __fill_dest_dir_rpm(
+        self,
+        game: GameData,
+        package: Package,
+        workdir: str,
+        destdir: str,
+        compress: Compression,
+        architecture: str,
+        release: str,
+    ) -> str:
         specfile = os.path.join(workdir, '%s.spec' % package.name)
         short_desc, long_desc = self.generate_description(game, package)
         short_desc = short_desc[0].upper() + short_desc[1:]
@@ -140,22 +172,22 @@ class RpmPackaging(PackagingSystem):
 
         files = set()
         for dirpath, dirnames, filenames in os.walk(destdir):
-             dir = dirpath[len(destdir):]
-             if not dir:
-                 # /
-                 continue
-             elif dir in SYSTEM_DIRS:
-                 for fn in filenames + dirnames:
-                     full = os.path.join(dirpath, fn)
-                     file = full[len(destdir):]
-                     if file not in SYSTEM_DIRS:
-                         files.add(file)
-             else:
-                 for file in files:
-                     if dir.startswith(file):
-                         break
-                 else:
-                     files.add(dir)
+            dir = dirpath[len(destdir):]
+            if not dir:
+                # /
+                continue
+            elif dir in SYSTEM_DIRS:
+                for fn in filenames + dirnames:
+                    full = os.path.join(dirpath, fn)
+                    file = full[len(destdir):]
+                    if file not in SYSTEM_DIRS:
+                        files.add(file)
+            else:
+                for file in files:
+                    if dir.startswith(file):
+                        break
+                else:
+                    files.add(dir)
 
         logger.debug('%%files in specfile:\n%s', '\n'.join(sorted(files)))
 
@@ -173,6 +205,10 @@ class RpmPackaging(PackagingSystem):
                 spec.write('Group: Amusements/Games\n')
             spec.write('BuildArch: %s\n' % architecture)
 
+            if package.mutually_exclusive:
+                for p in package.demo_for | package.better_versions:
+                    spec.write('Conflicts: %s\n' % p)
+
             for p in self.merge_relations(package, 'provides'):
                 spec.write('Provides: %s\n' % p)
 
@@ -192,8 +228,10 @@ class RpmPackaging(PackagingSystem):
             for p in self.merge_relations(package, 'depends'):
                 spec.write('Requires: %s\n' % p)
 
-            for p in (self.merge_relations(package, 'conflicts') |
-                    self.merge_relations(package, 'breaks')):
+            for p in (
+                self.merge_relations(package, 'conflicts')
+                | self.merge_relations(package, 'breaks')
+            ):
                 spec.write('Conflicts: %s\n' % p)
 
             for p in self.merge_relations(package, 'recommends'):
@@ -207,12 +245,12 @@ class RpmPackaging(PackagingSystem):
 
             # FIXME: replaces?
 
-            if not compress or package.rip_cd:
+            if not compress:
                 spec.write('%define _binary_payload w0.gzdio\n')
             elif compress == ['-Zgzip', '-z1']:
                 spec.write('%define _binary_payload w1.gzdio\n')
             spec.write('%description\n')
-            spec.write('%s\n' % long_desc)
+            spec.write('%s\n' % '\n'.join(long_desc))
             spec.write('%files\n')
             spec.write('\n'.join(files))
             spec.write('\n\n')
@@ -220,19 +258,32 @@ class RpmPackaging(PackagingSystem):
             spec.write('%changelog\n')
             try:
                 login = os.getlogin()
-            except FileNotFoundError:
+            except (FileNotFoundError, OSError):
                 login = 'game-data-packager'
-            spec.write('* %s %s@%s - %s-%s\n' %
-                        (time.strftime("%a %b %d %Y", time.gmtime()),
-                         login, os.uname()[1], package.version, release))
+            spec.write(
+                '* %s %s@%s - %s-%s\n' % (
+                    time.strftime("%a %b %d %Y", time.gmtime()),
+                    login,
+                    os.uname()[1],
+                    package.version,
+                    release,
+                )
+            )
             spec.write('- Package generated by game-data-packager'
                        ' for local use only\n')
 
         return specfile
 
-    def build_package(self, per_package_dir, game, package, destination,
-            compress=True, md5sums=None, component=None):
-        destdir = os.path.join(per_package_dir, 'DESTDIR')
+    def build_package(
+        self,
+        per_package_state: PerPackageState,
+        game: GameData,
+        destination: str,
+        compress: Compression = True,
+    ) -> str:
+        destdir = per_package_state.destdir
+        package = per_package_state.package
+        per_package_dir = per_package_state.per_package_dir
         arch = self.get_effective_architecture(package)
 
         if arch == 'noarch':
@@ -243,14 +294,17 @@ class RpmPackaging(PackagingSystem):
         # increase local 'release' number on repacking
         if not self.is_installed(package.name):
             release = '0'
-        elif Version(package.version) > Version(self.current_version(package.name)):
+        elif package.version != self.current_version(package.name):
             release = '0'
         else:
             try:
-                release = check_output(['rpm', '-q', '--qf' ,'%{RELEASE}',
-                                         package.name]).decode('ascii')
-                if (self.distro is not None and
-                        release.endswith('.' + self.distro)):
+                release = check_output(
+                    ['rpm', '-q', '--qf', '%{RELEASE}', package.name]
+                ).decode('ascii')
+                if (
+                    self.distro is not None
+                    and release.endswith('.' + self.distro)
+                ):
                     release = release[:-(len(self.distro) + 1)]
                 release = str(int(release) + 1)
             except (subprocess.CalledProcessError, ValueError):
@@ -259,28 +313,38 @@ class RpmPackaging(PackagingSystem):
         if self.distro is not None:
             release = release + '.' + self.distro
 
-        if compress:
-            compress = game.compress_deb
-
-        specfile = self.__fill_dest_dir_rpm(game, package,
-                per_package_dir, destdir, compress, arch, release)
+        specfile = self.__fill_dest_dir_rpm(
+            game, package, per_package_dir, destdir, compress, arch, release,
+        )
         normalize_permissions(destdir)
 
         assert os.path.isdir(os.path.join(destdir, 'usr')), destdir
 
         try:
             logger.info('generating package %s', package.name)
-            check_output(setarch  + ['rpmbuild',
-                         '--buildroot', destdir,
-                         '-bb', '-v', specfile],
-                         cwd=per_package_dir)
+            check_output(
+                setarch + [
+                    'rpmbuild', '--buildroot', destdir, '-bb', '-v', specfile,
+                ],
+                cwd=per_package_dir,
+            )
         except subprocess.CalledProcessError as cpe:
             print(cpe.output)
             raise
 
-        return(os.path.expanduser('~/rpmbuild/RPMS/') + arch + '/'
-                + package.name + '-'
-                + package.version + '-' + release + '.' + arch + '.rpm')
+        return (
+            os.path.expanduser('~/rpmbuild/RPMS/') + arch + '/'
+            + package.name + '-'
+            + package.version + '-' + release + '.' + arch + '.rpm'
+        )
+
+    def get_libdir(self, package_arch: str) -> str:
+        if '64' in self.get_architecture(package_arch):
+            # x86_64, aarch64, ppc64le, riscv64
+            return '/usr/lib64'
+        else:
+            return '/usr/lib'
+
 
 # XXX: dnf is written in python3 and has a stable public api,
 #      it is likely faster to use it instead of calling 'dnf' pgm.
@@ -291,102 +355,147 @@ class RpmPackaging(PackagingSystem):
 #      https://github.com/timlau/dnf-apiex
 #
 #      As install_packages() needs root, we need to use the 'dnf' pgm
-
 class DnfPackaging(RpmPackaging):
     LICENSEDIR = '/usr/share/licenses'
     INSTALL_CMD = ['dnf', 'install']
     PACKAGE_MAP = {
-                  'dpkg-deb': 'dpkg',
-                  'id-shr-extract': None,
-                  '7z': 'p7zip-plugins',
-                  'unrar-nonfree': 'unrar',
-                  }
+        'dpkg-deb': 'dpkg',
+        'id-shr-extract': None,
+        '7z': 'p7zip-plugins',
+        'unrar-nonfree': 'unrar',
+        'xdelta3': 'xdelta',
+    }
 
-    def __init__(self, distro='fedora'):
-        super(DnfPackaging, self).__init__(distro)
-        self.available = None
+    def __init__(
+        self,
+        distro: str = 'fedora',
+        architecture: str | None = None,
+    ) -> None:
+        super(DnfPackaging, self).__init__(distro, architecture)
+        self.available: set[str] | None = None
 
-    def read_architecture(self):
-        super(DnfPackaging, self).read_architecture()
-        if self._architecture == 'amd64':
+    def read_architecture(self) -> str:
+        primary = super(DnfPackaging, self).read_architecture()
+
+        if primary == 'amd64':
             self._foreign_architectures = set(['i386'])
 
-    def is_available(self, package):
+        return primary
+
+    def is_available(self, package: str) -> bool:
         if self.available is None:
             try:
-                proc = subprocess.Popen(['dnf', 'list'],
-                        universal_newlines=True,
-                        stderr=subprocess.DEVNULL,
-                        stdout=subprocess.PIPE)
+                proc = subprocess.Popen(
+                    ['dnf', 'list'],
+                    text=True,
+                    stderr=subprocess.DEVNULL,
+                    stdout=subprocess.PIPE,
+                )
             except FileNotFoundError:
                 return False
-            cache = set()
-            for line in proc.stdout:
+            cache: set[str] = set()
+            stdout = proc.stdout
+            assert stdout is not None
+            for line in stdout:
                 if '.' in line:
                     cache.add(line.split('.')[0])
             self.available = cache
 
         return package in self.available
 
-    def available_version(self, package):
+    def available_version(self, package: str) -> str | None:
         try:
-            proc = subprocess.Popen(['dnf', 'list', package],
-                                     universal_newlines=True,
-                                     stderr=subprocess.DEVNULL,
-                                     stdout=subprocess.PIPE)
+            proc = subprocess.Popen(
+                ['dnf', 'list', package],
+                text=True,
+                stderr=subprocess.DEVNULL,
+                stdout=subprocess.PIPE,
+            )
         except FileNotFoundError:
             return None
         # keep only last line
-        for line in proc.stdout:
+        stdout = proc.stdout
+        assert stdout is not None
+        for line in stdout:
             pass
         return line.split()[1]
 
-    def install_packages(self, rpms, method='dnf', gain_root='su'):
-        super(DnfPackaging, self).install_packages(rpms, method=method,
-                gain_root=gain_root)
+    def install_packages(
+        self,
+        rpms: Iterable[str],
+        method: str | None = 'dnf',
+        gain_root: str = 'su',
+        force: bool = False,
+    ) -> None:
+        super(DnfPackaging, self).install_packages(
+            rpms, method=method, gain_root=gain_root,
+        )
+
 
 class ZypperPackaging(RpmPackaging):
     DOCDIR = '/usr/share/doc/packages'
     LICENSEDIR = '/usr/share/doc/packages'
     INSTALL_CMD = ['zypper', 'install']
     PACKAGE_MAP = {
-                  'dpkg-deb': 'dpkg',
-                  'id-shr-extract': None,
-                  '7z': 'p7zip',
-                  'unrar-nonfree': 'unrar',
-                  }
+        'dpkg-deb': 'dpkg',
+        'id-shr-extract': None,
+        '7z': 'p7zip',
+        'unrar-nonfree': 'unrar',
+    }
 
-    def __init__(self, distro='suse'):
-        super(ZypperPackaging, self).__init__(distro)
+    def __init__(
+        self,
+        distro: str = 'suse',
+        architecture: str | None = None
+    ) -> None:
+        super(ZypperPackaging, self).__init__(distro, architecture)
 
-    def is_available(self, package):
+    def is_available(self, package: str) -> bool:
         try:
-            proc = subprocess.Popen(['zypper', 'info', package],
-                    universal_newlines=True,
-                    stdout=subprocess.PIPE,
-                    env={'LANG':'C'})
+            proc = subprocess.Popen(
+                ['zypper', 'info', package],
+                text=True,
+                stdout=subprocess.PIPE,
+                env={'LANG': 'C'},
+            )
         except FileNotFoundError:
             return False
-        for line in proc.stdout:
+        stdout = proc.stdout
+        assert stdout is not None
+        for line in stdout:
             if line.startswith('Version:'):
                 return True
         return False
 
-    def available_version(self, package):
+    def available_version(self, package: str) -> str | None:
         try:
-            proc = subprocess.Popen(['zypper', 'info', package],
-                    universal_newlines=True,
-                    stdout=subprocess.PIPE,
-                    env={'LANG':'C'})
+            proc = subprocess.Popen(
+                ['zypper', 'info', package],
+                text=True,
+                stdout=subprocess.PIPE,
+                env={'LANG': 'C'},
+            )
         except FileNotFoundError:
             return None
-        for line in proc.stdout:
+        stdout = proc.stdout
+        assert stdout is not None
+        for line in stdout:
             if line.startswith('Version:'):
                 return line.split(':', maxsplit=1)[1]
 
-    def install_packages(self, rpms, method='zypper', gain_root='su'):
-        super(ZypperPackaging, self).install_packages(rpms, method=method,
-                gain_root=gain_root)
+        return None
+
+    def install_packages(
+        self,
+        rpms: Iterable[str],
+        method: str | None = 'zypper',
+        gain_root: str = 'su',
+        force: bool = False,
+    ) -> None:
+        super(ZypperPackaging, self).install_packages(
+            rpms, method=method, gain_root=gain_root,
+        )
+
 
 class UrpmiPackaging(RpmPackaging):
     BINDIR = '/usr/games'
@@ -399,40 +508,52 @@ class UrpmiPackaging(RpmPackaging):
                   'unrar-nonfree': 'unrar',
                   }
 
-    def __init__(self, distro='mageia'):
-        super(UrpmiPackaging, self).__init__(distro)
+    def __init__(
+        self,
+        distro: str = 'mageia',
+        architecture: str | None = None,
+    ) -> None:
+        super(UrpmiPackaging, self).__init__(distro, architecture)
 
-    def is_available(self, package):
+    def is_available(self, package: str) -> bool:
         try:
-            return 0 == subprocess.call(['urpmq', package],
-                                        stdout=subprocess.DEVNULL,
-                                        stderr=subprocess.DEVNULL)
+            return 0 == subprocess.call(
+                ['urpmq', package],
+                stdout=subprocess.DEVNULL,
+                stderr=subprocess.DEVNULL,
+            )
         except FileNotFoundError:
             return False
 
-    def available_version(self, package):
+    def available_version(self, package: str) -> str | None:
         try:
             line = check_output(['urpmq', '-r', package]).decode('ascii')
+            assert type(line) is str
             return line.split('-')[-2]
         except FileNotFoundError:
             return None
         except subprocess.CalledProcessError:
             return None
 
-def get_packaging_system(distro=None):
+
+def get_packaging_system(
+    distro: str | None = None,
+    architecture: str | None = None,
+) -> RpmPackaging:
     if distro is None:
         distro = _get_distro()
 
     if distro == 'mageia':
-        return UrpmiPackaging(distro)
+        return UrpmiPackaging(distro, architecture)
     elif distro == 'fedora':
-        return DnfPackaging(distro)
+        return DnfPackaging(distro, architecture)
     elif distro == 'suse':
-        return ZypperPackaging(distro)
+        return ZypperPackaging(distro, architecture)
 
-    return RpmPackaging(distro)
+    return RpmPackaging(distro, architecture)
 
-def _get_distro():
+
+def _get_distro() -> str:
     if os.path.isfile('/etc/mageia-release'):
         return 'mageia'
 
@@ -443,11 +564,11 @@ def _get_distro():
         return 'suse'
 
     try:
-        maybe = DnfPackaging()
+        maybe: RpmPackaging = DnfPackaging()
 
         if maybe.is_available('rpm'):
             return 'fedora'
-    except:
+    except Exception:
         pass
 
     try:
@@ -455,7 +576,7 @@ def _get_distro():
 
         if maybe.is_available('rpm'):
             return 'suse'
-    except:
+    except Exception:
         pass
 
     return 'generic'
